#!/usr/bin/python3

"""WMI device information"""

from __future__ import annotations
from dataclasses import dataclass
from enum import StrEnum
from typing import Optional, Iterable
from pathlib import Path
from os import scandir


class WMIDeviceType(StrEnum):
    """Type of an WMI device"""
    DATABLOCK = "Data"
    METHOD = "Method"
    EVENT = "Event"


def wmi_bus_devices(bus_path: Path = Path('/sys/bus/wmi/devices')) -> Iterable[Path]:
    """Scan for WMI devices on the WMI bus"""
    try:
        with scandir(bus_path) as directory:
            for entry in directory:
                if entry.is_dir():
                    yield Path(entry.path)
    except FileNotFoundError:
        pass


@dataclass(frozen=True, slots=True)
class WMIDevice:
    """WMI device information"""

    guid: str

    device_type: WMIDeviceType

    device_id: str

    instances: int

    expensive: bool

    setable: Optional[bool]

    driver: Optional[str]

    path: Path

    @classmethod
    def from_sysfs(cls, path: Path) -> WMIDevice:
        """Parse WMI device information from sysfs"""
        setable = None

        with (path / 'guid').open('r', encoding='ascii') as fd:
            guid = fd.read().rstrip()

        with (path / 'instance_count').open('r', encoding='ascii') as fd:
            instances = int(fd.read())

        with (path / 'expensive').open('r', encoding='ascii') as fd:
            expensive = int(fd.read()) != 0

        try:
            driver = (path / 'driver').readlink().name
        except FileNotFoundError:
            driver = None

        try:
            with (path / 'notify_id').open('r', encoding='ascii') as fd:
                device_id = fd.read().rstrip()
                device_type = WMIDeviceType.EVENT
        except FileNotFoundError:
            with (path / 'object_id').open('r', encoding='ascii') as fd:
                device_id = fd.read().rstrip()

            try:
                with (path / 'setable').open('r', encoding='ascii') as fd:
                    setable = int(fd.read()) != 0
                    device_type = WMIDeviceType.DATABLOCK
            except FileNotFoundError:
                device_type = WMIDeviceType.METHOD

        return cls(
            guid=guid,
            device_type=device_type,
            device_id=device_id,
            instances=instances,
            expensive=expensive,
            setable=setable,
            driver=driver,
            path=path
        )



#!/usr/bin/python3

"""Utility to retrieve information about WMI devices on Linux """

__version__ = "1.1.0"
__author__ = "Armin Wolf"
__email__ = "W_Armin@gmx.de"
__all__ = (
    "device",
    "provider",
    "main"
)



#!/usr/bin/python3

"""CLI entrypoint utilities"""

from argparse import ArgumentParser, Namespace
from typing import Final, Iterable
from . import __doc__ as description, __version__
from .device import WMIDevice, wmi_bus_devices

__all__ = (
    "ARGUMENT_PARSER",
    "main",
    "main_cli"
)

ARGUMENT_PARSER: Final = ArgumentParser(
    prog="lswmi",
    description=description
)
ARGUMENT_PARSER.add_argument(
    "-v",
    "--version",
    action="version",
    version=f"%(prog)s {__version__}"
)
ARGUMENT_PARSER.add_argument(
    "-V",
    "--verbose",
    action="store_true",
    help="enable verbose output"
)
ARGUMENT_PARSER.add_argument(
    "-g",
    "--guid",
    action="store",
    help="show only devices with the specified GUID"
)


def print_device_summary(device: WMIDevice) -> None:
    """Print a short WMI device summary"""
    print(f'{device.guid}: {device.device_type:<6} (Instances: {device.instances})')


def print_device_verbose(device: WMIDevice) -> None:
    """Print additional WMI device information"""
    print(f'    Identification: {device.device_id}')
    print(f'    Expensive: {device.expensive}')
    if device.setable is not None:
        print(f'    Setable: {device.setable}')

    if device.driver is not None:
        print(f'    Driver: {device.driver}')


def wmi_devices() -> Iterable[WMIDevice]:
    """Scan for WMI devices"""
    for path in wmi_bus_devices():
        try:
            device = WMIDevice.from_sysfs(path)
        except OSError as error:
            print(f'Unable to read WMI device: {error}')
            continue

        yield device


def main(args: Namespace) -> int:
    """Entrypoint for the ePPID tool"""
    devices = wmi_devices()
    if args.guid is not None:
        devices = filter(lambda d: d.guid == args.guid, devices)

    for device in devices:
        print_device_summary(device)
        if args.verbose:
            print_device_verbose(device)
            print('')

    return 0


def main_cli() -> int:
    """CLI entry point"""
    return main(ARGUMENT_PARSER.parse_args())



#!/usr/bin/python3

"""Test for WMI device handling"""

from unittest import TestCase
from pathlib import Path
from tempfile import TemporaryDirectory
from lswmi.device import WMIDevice, WMIDeviceType, wmi_bus_devices


def create_file(path: Path, content: str) -> None:
    """Create file with specific content"""
    with path.open('x', encoding='ascii') as fd:
        fd.write(f'{content}\n')


def create_wmi_device(device: WMIDevice) -> None:
    """Create WMI device sysfs structure"""
    device_dir = Path(device.path)
    device_dir.mkdir(mode=0o755)

    create_file((device_dir / 'guid'), device.guid)
    create_file((device_dir / 'expensive'), str(int(device.expensive)))
    create_file((device_dir / 'instance_count'), str(device.instances))

    if device.device_type == WMIDeviceType.EVENT:
        create_file((device_dir / 'notify_id'), device.device_id)
    else:
        create_file((device_dir / 'object_id'), device.device_id)

    if device.setable is not None:
        create_file((device_dir / 'setable'), str(int(device.setable)))

    if device.driver is not None:
        (device_dir / 'driver').symlink_to(f'../../drivers/{device.driver}')


class WmiDeviceTest(TestCase):
    """Tests for WMI device handling"""

    def setUp(self) -> None:
        """Prepare fake WMI bus directory"""
        bus_path = Path(self.enterContext(TemporaryDirectory()))
        self.device_path = bus_path / 'devices'
        self.device_path.mkdir(mode=0o755)

        self.bmof_device = WMIDevice(
            guid='05901221-D566-11D1-B2F0-00A0C9062910',
            device_type=WMIDeviceType.DATABLOCK,
            device_id='MO',
            instances=1,
            expensive=False,
            setable=False,
            driver='wmi-bmof',
            path=(self.device_path / '05901221-D566-11D1-B2F0-00A0C9062910')
        )
        create_wmi_device(self.bmof_device)

        self.bmof_device_duplicate = WMIDevice(
            guid='05901221-D566-11D1-B2F0-00A0C9062910',
            device_type=WMIDeviceType.DATABLOCK,
            device_id='MO',
            instances=1,
            expensive=False,
            setable=False,
            driver='wmi-bmof',
            path=(self.device_path / '05901221-D566-11D1-B2F0-00A0C9062910-1')
        )
        create_wmi_device(self.bmof_device_duplicate)

        self.ddv_device = WMIDevice(
            guid='8A42EA14-4F2A-FD45-6422-0087F7A7E608',
            device_type=WMIDeviceType.METHOD,
            device_id='DV',
            instances=1,
            expensive=False,
            setable=None,
            driver='dell-wmi-ddv',
            path=(self.device_path / '8A42EA14-4F2A-FD45-6422-0087F7A7E608')
        )
        create_wmi_device(self.ddv_device)

        self.eeepc_device = WMIDevice(
            guid='ABBC0F72-8EA1-11D1-00A0-C90629100000',
            device_type=WMIDeviceType.EVENT,
            device_id='D2',
            instances=1,
            expensive=False,
            setable=None,
            driver=None,
            path=(self.device_path / 'ABBC0F72-8EA1-11D1-00A0-C90629100000')
        )
        create_wmi_device(self.eeepc_device)

    def test_simple_datablock(self) -> None:
        """Test detection of an simple WMI datablock device"""
        device = WMIDevice.from_sysfs(
            self.device_path / '05901221-D566-11D1-B2F0-00A0C9062910'
        )

        self.assertEqual(device, self.bmof_device)

    def test_simple_method(self) -> None:
        """Test detection of an simple WMI method device"""
        device = WMIDevice.from_sysfs(
            self.device_path / '8A42EA14-4F2A-FD45-6422-0087F7A7E608'
        )

        self.assertEqual(device, self.ddv_device)

    def test_simple_event(self) -> None:
        """Test detection of an simple WMI event device"""
        device = WMIDevice.from_sysfs(
            self.device_path / 'ABBC0F72-8EA1-11D1-00A0-C90629100000'
        )

        self.assertEqual(device, self.eeepc_device)

    def test_duplicate(self) -> None:
        """Test detection of an WMI device with an duplicate GUID"""
        device = WMIDevice.from_sysfs(
            self.device_path / '05901221-D566-11D1-B2F0-00A0C9062910-1'
        )

        self.assertEqual(device, self.bmof_device_duplicate)

    def test_detection(self) -> None:
        """Test detection of all WMI devices"""
        devices = list(wmi_bus_devices(self.device_path))

        self.assertIn(self.bmof_device.path, devices)
        self.assertIn(self.bmof_device_duplicate.path, devices)
        self.assertIn(self.ddv_device.path, devices)
        self.assertIn(self.eeepc_device.path, devices)
        self.assertEqual(len(devices), 4)

    def test_detection_nondir(self) -> None:
        """Test if device detection ignores files"""
        create_file(
            self.device_path / '05901221-D566-11D1-B2F0-00A0C9062910-2',
            'dummy'
        )

        with self.subTest("Device detection did not ignore file"):
            self.test_detection()

    def test_detection_nonbus(self) -> None:
        """Test if device detection handles absence of the WMI bus"""
        devices = list(wmi_bus_devices(self.device_path / 'fake'))

        self.assertEqual(len(devices), 0)
